(******************************************************************
Hash map

Copyright (C) 2005, 07 by raptor.zh@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

website: http://mental.we8log.com
blog   : http://blog.csdn.net/raptor
Rev.   : Mar.05 Mar.13

IMPair: MapEntryӿڡ
TMIntegerKeyPair: һΪKeyPair
TMGUIDKeyPair: һGUIDΪKeyPair
TMSStrKeyPair: һShortStringΪKeyPair
TMHashMap: Hash mapṹ

Apr.24 IndexOfرRange checking

Sep.22 TMSStrKeyPair
******************************************************************)
unit MHashMap;

interface

Uses Classes, SysUtils;

const
  cnHashSize = 256;

type
  TMHashMap = class;

  TMHashType = Byte;
  TMValueObject = TObject;

  IMPair = interface
    function HashKey : TMHashType;
    function IsEqualKey( aPair : IMPair ) : Boolean;
    function GetValue : TMValueObject;
    function GetObject : TObject;

    property Value : TMValueObject read GetValue;
  end;

  TMCustomPair = class(TInterfacedObject, IMPair)
  private
    FValue : TObject;

    function HashKey : TMHashType; Virtual; Abstract;
    function IsEqualKey( aPair : IMPair ) : Boolean; Virtual; Abstract;
    function GetValue : TMValueObject;
    function GetObject : TObject;
  public
    Constructor Create( aValue : TMValueObject );
    Destructor Destroy; Override;
  End;

  TMIntegerKeyPair = class(TMCustomPair)
  protected
    FKey   : Integer;

    function HashKey : TMHashType; Override;
    function IsEqualKey( aPair : IMPair ) : Boolean; Override;
  public
    constructor Create( aKey : Integer; aValue : TMValueObject );

    property Key : Integer read FKey;
  End;

  TMGUIDKeyPair = class(TMCustomPair)
  protected
    FKey   : TGUID;

    function HashKey : TMHashType; Override;
    function IsEqualKey( aPair : IMPair ) : Boolean; Override;
  public
    constructor Create( aKey : TGUID; aValue : TMValueObject );

    property Key : TGUID read FKey;
  End;

  // TODO: δԣݲá
  TMSStrKeyPair = class(TMCustomPair)
  protected
    FKey   : ShortString;

    function HashKey : TMHashType; Override;
    function IsEqualKey( aPair : IMPair ) : Boolean; Override;
  public
    constructor Create( aKey : ShortString; aValue : TMValueObject );

    property Key : ShortString read FKey;
  End;

  TMHashMap = class(TObject)
  private
    FData : array [0..cnHashSize-1] of IMPair;
  protected
    function  FindEmpty( aIndex : TMHashType ) : TMHashType;
    function  IndexOf( aPair : IMPair ) : TMHashType;
  public
    destructor Destroy; Override;

    procedure Add( aPair : IMPair );
    procedure Delete( aPair : IMPair );
    procedure Clear( );
    function  Find( aPair : IMPair ) : IMPair;
  End;

implementation

resourcestring
    SMapFull      = 'Map is full!';
    SPairNotFound = 'Pair not found!';

{ TMCustomPair }

constructor TMCustomPair.Create(aValue: TMValueObject);
begin
    FValue := aValue;
end;

destructor TMCustomPair.Destroy;
begin
    If Assigned( FValue ) Then
        FValue.Free;    
    inherited;
end;

function TMCustomPair.GetValue: TMValueObject;
begin
    Result := FValue;
end;

function TMCustomPair.GetObject: TObject;
begin
    Result := Self;
end;

{ TMIntegerKeyPair }

constructor TMIntegerKeyPair.Create(aKey: Integer; aValue: TMValueObject);
begin
    FKey := aKey;
    inherited Create( aValue );
end;

function TMIntegerKeyPair.HashKey: TMHashType;
begin
    Result := TMHashType( FKey );  //  simple hash
end;

function TMIntegerKeyPair.IsEqualKey(aPair: IMPair): Boolean;
begin
    Result := ( aPair.GetObject Is TMIntegerKeyPair )
        AND ( ( aPair.GetObject As TMIntegerKeyPair ).FKey = FKey );
end;

{ TMGUIDKeyPair }

constructor TMGUIDKeyPair.Create(aKey: TGUID; aValue: TMValueObject);
begin
    FKey := aKey;
    inherited Create( aValue );
end;

function TMGUIDKeyPair.HashKey: TMHashType;
begin
    Result := TMHashType( FKey.D1 );  //  simple hash
end;

function TMGUIDKeyPair.IsEqualKey( aPair : IMPair ): Boolean;
begin
    Result := ( aPair.GetObject Is TMGUIDKeyPair )
        AND IsEqualGUID( ( aPair.GetObject As TMGUIDKeyPair ).FKey, FKey );
end;

{ TMSStrKeyPair }

constructor TMSStrKeyPair.Create(aKey: ShortString; aValue: TMValueObject);
begin
    FKey := aKey;
    inherited Create( aValue );
end;

function TMSStrKeyPair.HashKey: TMHashType;
Var
  i, n : Integer;
begin
  n := 0;
  For i := Length( FKey ) DownTo 1 Do
    n := n + Ord( FKey[i] );
  Result := TMHashType( n );
end;

function TMSStrKeyPair.IsEqualKey(aPair: IMPair): Boolean;
begin
    Result := ( aPair.GetObject Is TMSStrKeyPair )
        AND ( ( aPair.GetObject As TMSStrKeyPair ).FKey = FKey );
end;

{ TMHashMap }

destructor TMHashMap.Destroy;
begin
  Clear();
  inherited;
end;

function TMHashMap.FindEmpty(aIndex: TMHashType): TMHashType;
begin
    Result := aIndex;
    While Assigned( FData[Result] ) Do
    Begin
        Inc( Result );
        If ( Result = aIndex ) Then
            Raise Exception.Create( SMapFull );
    End;
end;

function TMHashMap.IndexOf(aPair: IMPair): TMHashType;
Var
    n : TMHashType;
begin
    n := aPair.HashKey( );
    Result := n;
{$R-}
    While ( Not Assigned( FData[Result] ) )
        OR ( Not FData[Result].IsEqualKey( aPair ) ) Do
    Begin
        Inc( Result );
        If ( Result = n ) Then
            Raise Exception.Create( SPairNotFound );
    End;
{$R+}
end;

procedure TMHashMap.Add(aPair: IMPair);
Var
    i : TMHashType;
begin
    i := aPair.HashKey();
    i := FindEmpty( i );
    FData[i] := aPair;
end;

procedure TMHashMap.Clear;
Var
    i : TMHashType;
    p : IMPair;
begin
    For i := 0 To cnHashSize - 1 Do
    Begin
        p := FData[i];
        If Assigned( p ) Then
            FData[i] := Nil;
    End;
end;

procedure TMHashMap.Delete(aPair: IMPair);
Var
    i : TMHashType;
begin
    Try
        i := IndexOf( aPair );
        FData[i] := Nil;
    Except
        //  do nothing when this key is not found
    End;
end;

function TMHashMap.Find(aPair: IMPair): IMPair;
begin
    Try
        Result := FData[IndexOf( aPair )];
    Except
        Result := Nil;
    End;
end;

end.
